# ----------------------------------------------------------------------------
#
#  Copyright (C) 2015 Fons Adriaensen <fons@linuxaudio.org>
#    
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
# ----------------------------------------------------------------------------


import alsacontrol_ext


class AlsaControl() :

    """
    This class provides access to ALSA's low level control
    interface. The functionality is roughly that of the
    amixer 'cget' and 'cset' commands when the control
    element is selected using its 'numid'.
    The element type IEC958 is currently not supported
    since its semantics are unclear.
    """

    
    TYPE_BOOL   = 1
    TYPE_INT    = 2
    TYPE_ENUM   = 3
    TYPE_BYTES  = 4
    TYPE_IEC958 = 5
    TYPE_INT64  = 6

    ERR_NONE   =  0
    ERR_DEVICE = -1
    ERR_NUMID  = -2
    ERR_ACCESS = -3
    ERR_READ   = -4
    ERR_WRITE  = -5
    ERR_TYPE   = -6

    
    def __init__(self):
        """
        Create a new AlsaControl instance.
        """
        self._capsule = alsacontrol_ext.capsule ((self,))


    def open (self, device) :
        """
        Open the control interface of the given device.
        Returns ERR_DEVICE if failed, zero otherwise.
        """
        return alsacontrol_ext.open ((self._capsule, device))


    def close (self) :
        """
        Close the current control interface.
        """
        return alsacontrol_ext.close ((self._capsule,))


    def cget(self, numid) :
        """
        Get the current value(s) for the element identified
        by 'numid'. Returns a tuple consisting of
          - the element type code, or a negative error code,
          - a list of integer values,
          - the min and max values for this element. 
        """
        return alsacontrol_ext.cget ((self._capsule, numid))


    def cset(self, numid, param) :
        """
        Set the value(s) for the element identified by 'numid'.
        'param' should be a list of values to be set.
        Returns the number of values actually set, or a negative
        error code. 
        """
        return alsacontrol_ext.cset ((self._capsule, numid, param))


